{-----------------------------------------------------------------------------
 Unit Name: YDPDictionary
 Author:    Maras
 Version:   v 0.31
 Purpose:   Component for access YDP dictionary files. Working w/ TRitchEdit
            and TListBox control if avaliable.
            Usage:
            property DictionaryFile - dictionaty file (*.idx or *.dat) both
              files must be present for working.
            property RichTextControl - optional property. If present component
              automaticly updates control w/ translation after seek. Remember
              to take care about PlainText property.
            property ListBoxControl - optional property. If present component
              automaticly updates world list after index loading.

            calling:
            - LoadIndex
            - oprional (assign FileIndex to ListBox.Items)
            - SeekWord
            - optional (load TranslationStream to RichText.Items)
            - to get rtf translation code get TranslationStream.DataString
 History:
 2003-05-08   First working version. No known problems. Need time to implement
            additional features.
 2003-05-23   Added colouring & some code optimalization. Better word seeking.
            Property DictionaryName added.
 2003-06-04   TWordIndexList class defined. Helps create user interface. Creates
            word indeks of file loaded (linne added etc.)

 Contact:     marek_74@wp.pl
-----------------------------------------------------------------------------}


unit YDPDictionary;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls,
  stdctrls;

type
  TYDPDictionary = class(TComponent)
  private
    FDictionaryFile: String;
    FRichTextControl: TRichEdit;
    FListBoxControl: TListBox;
    FDictionaryName: String;
    procedure SetDictionaryFile(const Value: String);
    procedure SetRichTextControl(const Value: TRichEdit);
    procedure SetListBoxControl(const Value: TListBox);
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }

{ FileIndex - stores dictionary index. Strings[i] stores word, Objects[i] stores
  file position. }
    FileIndex : TStringList;
{ TranslationStream - stores translation after SeekWord. You can load from this
  stream exactly to control (TRichText) or get data by object property (watch
  Delphi help). }
    TranslationStream : TStringStream;

    constructor Create(Owner : TComponent);override;
    constructor Create_(DictionaryFile : String);virtual;
{-----------------------------------------------------------------------------
  Function: LoadIndex
  Loads index file into component. After run it fills FileIndex property &
  ListBox control if given. It gives number of words read as Result.
-----------------------------------------------------------------------------}
    function LoadIndex:Integer;
{-----------------------------------------------------------------------------
  Function: SeekWord
  Seeks to word (translation) given by index number. After run it fills
  TranslationStream property & RitchTextControl if given. It gives seeked word
  as Result.
-----------------------------------------------------------------------------}
    function SeekWord( Index : Integer) : String;overload;
{-----------------------------------------------------------------------------
  Function: SeekWord
  The same like above, but for given word.
-----------------------------------------------------------------------------}
    function SeekWord( Word : String) : Integer;overload;

  published
    { Published declarations }
{ You have to put index filename here}
    property DictionaryFile : String read FDictionaryFile write SetDictionaryFile;
{ You can give TRichText control here. When present it is automaticly filled
  w/ translation tekst.}
    property RichTextControl : TRichEdit read FRichTextControl write SetRichTextControl;
{ You can give TListBox control here. When present it is automaticly filled
  w/ index word list. }
    property ListBoxControl : TListBox read FListBoxControl write SetListBoxControl;
{ Read only property. Dictionary name is avaliable there}
    property DictionaryName : String read FDictionaryName;
  end;



  TWordIndexList = class(TStringList)
  private
    FLowCase: Boolean;
    FMinWordLength: Word;
    FAllowDigits: Boolean;
    procedure SetLowCase(const Value: Boolean);
    procedure SetMinWordLength(const Value: Word);
    procedure SetAllowDigits(const Value: Boolean);
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create;
    function Add(const Value: string): Integer; override;
  published
    { Published declarations }
{ Convert all words to locase ? }
    property LowCase : Boolean read FLowCase write SetLowCase;
{ Determines what minimal word length will be added to index }
    property MinWordLength : Word read FMinWordLength write SetMinWordLength;
{ Digit is wvalid character ? }
    property AllowDigits : Boolean read FAllowDigits write SetAllowDigits;
  end;



procedure Register;

const

//FileSeek mode constans
  OFFSET_BEGIN = 0;
  OFFSET_CURRENT = 1;
  OFFSER_END = 2;

implementation

procedure Register;
begin
  RegisterComponents('Maras', [TYDPDictionary]);
end;


{ TYDPDictionary }
const
  // f0 - czcionka podstawowa
  // f1 - czcionka fonetyczna
  // f2 - czcionka tumaczenia
//  FontDef ='{\rtf1\ansi\ansicpg1250\deff0\deflang1045{\fonttbl{\f0\fswiss\fcharset238{\*\fname Arial;}Arial CE;}{\f1\fswiss\fprq2\fcharset238 Arial Narrow;}{\f2\fswiss\fprq2\fcharset238 Tahoma;}}';
//{\f5\fswiss\fcharset0 Arial;}
  FontDef ='{\rtf1\ansi\ansicpg1250\deff0\deflang1045{\fonttbl{\f0\fswiss\fcharset238 Arial;}{\f1\fswiss\fcharset0 Phonetic;}{\f2\fswiss\fcharset238 Tahoma;}}';
  ColorDef = '{\colortbl \red0\green128\blue0;\red128\green0\blue0;\red0\green0\blue128;\red128\green128\blue0;\red128\green0\blue128;\red0\green128\blue128;\red64\green64\blue64;}';

procedure TYDPDictionary.SetDictionaryFile(const Value: String);
begin
  FDictionaryFile := Value;
end;

constructor TYDPDictionary.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FileIndex := TStringList.Create;
  FileIndex.Duplicates := dupAccept;
  TranslationStream := TStringStream.Create('');
end;

constructor TYDPDictionary.Create_(DictionaryFile: String);
begin
  Create(nil);
  SetDictionaryFile(DictionaryFile);
end;

function TYDPDictionary.LoadIndex: Integer;
var
  LIndexName : String;
  LIndexFile : Integer;
  LIndexLength : Integer;
  LIndexRead : Integer;
  Buffer  : PChar;
  BufferT : PChar;
  tmpPossition : Longword;
  tmpWord : String;
  i : Integer;
  LFileBegin : Cardinal;

  LDexFileName : String;
  LDexFile : Integer;

const
  LFilePos = $10; //index offset stored here (i hope)
  LDexFilePos = 6; // dictionary name stored here
begin
  Result := 0;
  try
    FileIndex.Clear;
    FDictionaryName:='';
    LDexFileName := ChangeFileExt(FDictionaryFile,'.dex');
    LDexFile := FileOpen(LDexFileName,fmOpenRead);
    FileSeek(LDexFile,LDexFilePos,OFFSET_BEGIN);
    Buffer := PChar(AllocMem(256));
    FileRead(LDexFile,Buffer^,255);
    FileClose(LDexFile);
    Buffer[255] := #0;
    FDictionaryName := String(Buffer);
    FreeMem(Buffer);

    LIndexName := ChangeFileExt(FDictionaryFile,'.idx');
    LIndexFile := FileOpen(LIndexName,fmOpenRead);
    LIndexLength := FileSeek(LIndexFile,0,OFFSER_END);
    FileSeek(LIndexFile,0,OFFSET_BEGIN);

    Buffer := PChar(AllocMem(LIndexLength + 1));
    LIndexRead := FileRead(LIndexFile, Buffer^, LIndexLength);
    FileClose(LIndexFile);
//    BufferI := PLongWord(Buffer + i);
    LFileBegin := PLongWord(Buffer +LFilePos)^;//BufferI^;
    i := LFileBegin + 4;
    while i < LIndexRead do
    begin
//      BufferI := PLongWord(Buffer + i);
      tmpPossition := PLongWord(Buffer + i)^;
      Inc(i, 4);
      BufferT := Buffer + i;
      tmpWord := String(BufferT);
      Inc(i, StrLen(BufferT) + 1 + 4);
      FileIndex.AddObject(tmpWord,TObject(tmpPossition));
      Result := FileIndex.Count;
    end;
    if FListBoxControl <> nil then
    begin
      FListBoxControl.Items.BeginUpdate;
      FListBoxControl.Items.Assign(FileIndex);
      FListBoxControl.Items.EndUpdate;
    end;
    except
      Result := -1;
    end;
    FreeMem(Buffer);

end;

function TYDPDictionary.SeekWord(Index: Integer): String;
var
  LDicPossition : Cardinal;
  LDicFileName : String;
  LDicFile : Integer;
  LDicRead : Integer;
  LTranslationSize : Integer;
  LBuffer : PChar;
begin
  if (Index < 0) or (Index > FileIndex.Count) then
    Exit;
  LDicPossition := Cardinal(FileIndex.Objects[Index]);
  LDicFileName := ChangeFileExt(DictionaryFile,'.dat');
  LDicFile := FileOpen(LDicFileName,fmOpenRead);
  FileSeek(LDicFile,LDicPossition,OFFSET_BEGIN);
  LTranslationSize := 0;
  FileRead(LDicFile,LTranslationSize,4);
  try
    LBuffer := PChar(AllocMem(LTranslationSize+1));
    LDicRead := FileRead(LDicFile, LBuffer^, LTranslationSize);
    FileClose(LDicFile);
    TranslationStream.Free;
    TranslationStream := TStringStream.Create('');
    TranslationStream.WriteString(FontDef);
    TranslationStream.WriteString(ColorDef);
    TranslationStream.Write(LBuffer^,LDicRead);
    TranslationStream.WriteString('}');
    TranslationStream.Seek(0,soFromBeginning);
    if FRichTextControl <> nil then
    begin
      FRichTextControl.Lines.BeginUpdate;
      FRichTextControl.Lines.LoadFromStream(TranslationStream);
      FRichTextControl.Lines.EndUpdate;
      TranslationStream.Seek(0,soFromBeginning);
    end;
    finally
      FreeMem(LBuffer);
    end;

end;

function TYDPDictionary.SeekWord(Word: String): Integer;
begin
  if FileIndex.Find(Word, Result) then
    SeekWord(Result)
  else
  begin
    if Result < 0 then
      Result := 0;
    SeekWord(Result)
  end;
end;

procedure TYDPDictionary.SetRichTextControl(const Value: TRichEdit);
begin
  FRichTextControl := Value;

  if FRichTextControl <> nil then
  begin
    FRichTextControl.FreeNotification(Self);
  end
end;

procedure TYDPDictionary.SetListBoxControl(const Value: TListBox);
begin
  FListBoxControl := Value;

  if FListBoxControl <> nil then
  begin
    FListBoxControl.FreeNotification(Self);
  end
end;



{ TWordIndexList }

function TWordIndexList.Add(const Value: string): Integer;
var
  P, P1: PChar;
  S : String;
begin
  if Value = '' then Exit;
  P := PChar(Value);
  if FLowCase then AnsiStrLower(P);
  while P^ <> #0 do
  begin
    if (FAllowDigits and IsCharAlphaNumeric(P^)) or (IsCharAlpha(P^)) then
    begin
      P1 := P;
      while (FAllowDigits and IsCharAlphaNumeric(P^)) or (IsCharAlpha(P^)) do
        P := CharNext(P);
      if (P-P1) >= FMinWordLength then
      begin
        SetString(S, P1, P - P1);
        inherited Add(S);
      end;
    end
    else
      P := CharNext(P);
  end;
end;

constructor TWordIndexList.Create;
begin
  inherited;
  Sorted := True;
  Duplicates := dupIgnore;
  FLowCase := True;
  FMinWordLength := 3;
  FAllowDigits := False;
end;

procedure TWordIndexList.SetAllowDigits(const Value: Boolean);
begin
  FAllowDigits := Value;
end;

procedure TWordIndexList.SetLowCase(const Value: Boolean);
begin
  FLowCase := Value;
end;

procedure TWordIndexList.SetMinWordLength(const Value: Word);
begin
  FMinWordLength := Value;
end;

end.
